# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import re
import time
import base64
import logging
import threading
import collections

class BaseSender:

    def __init__(self, parent, usb_info, profile):
        self.logger = logging.getLogger(__name__)
        self.stop_flag = False
        self.parent = parent
        self.profile = profile
        self.usb_info = usb_info
        self.position = [0, 0, 0, 0]  # X, Y, Z, E
        self.total_gcodes = None
        self.buffer = collections.deque()
        self.current_line_number = 0
        self.pause_flag = False
        self.callbacks_lock = threading.Lock()
        self.response_callbacks = [] # fuctions list to call on printer_response. if not empty it will suppress tempratures requesting.
        self.responses_planned = 0
        self.responses = []
        self.filename = None
        self.is_base64_re = re.compile(b"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$")
        if not getattr(self, "temps", False):
            self.temps = [0,0]
            self.target_temps = [0,0]

    def set_total_gcodes(self, length):
        raise NotImplementedError

    def load_gcodes(self, gcodes):
        raise NotImplementedError

    def unbuffered_gcodes(self, gcodes):
        raise NotImplementedError

    def cancel(self):
        self.parent.register_error(605, "Cancel is not supported for this printer type", is_blocking=False)
        return False

    def preprocess_gcodes(self, gcodes):
        gcodes = gcodes.replace(b"\r", b"")
        gcodes = gcodes.split(b"\n")
        gcodes = [item.decode("utf-8") for item in gcodes if item]
        if gcodes:
            while gcodes[-1] in ("\n", "\n", "\t", " ", "", None):
                line = gcodes.pop()
                self.logger.info("Removing corrupted line '%s' from gcodes tail" % line)
        self.logger.info('Got %d gcodes to print.' % len(gcodes))
        return gcodes

    def gcodes(self, gcodes):
        start_time = time.monotonic()
        self.logger.debug("Determining gcode format")
        if self.is_base64_re.match(gcodes):
            gcodes = base64.b64decode(gcodes)
            self.unbuffered_gcodes(gcodes)
        else:
            self.logger.debug("Start loading gcodes. Determination time:" + str(time.monotonic()-start_time))
            self.load_gcodes(gcodes)
            self.logger.debug("Done loading gcodes. Time:" + str(time.monotonic()-start_time))

    def set_filename(self, filename):
        self.filename = str(filename) if filename else None

    def get_position(self):
        return self.position

    def get_temps(self):
        return self.temps

    def get_target_temps(self):
        return self.target_temps

    def pause(self):
        self.pause_flag = True

    def unpause(self):
        self.pause_flag = False

    def close(self):
        self.stop_flag = True

    def is_paused(self):
        return self.pause_flag

    def is_operational(self):
        return False

    def get_downloading_percent(self):
        return self.parent.downloader.get_percent()

    def get_nonstandart_data(self):
        return {}

    def execute_callback(self, line, success):
        for callback in self.response_callbacks:
            try:
                callback(line, success)
            except:
                self.logger.exception("Exception in callback(%s):" % str(callback))

    def add_response_callback(self, callback_function):
        with self.callbacks_lock:
            self.response_callbacks.append(callback_function)

    def del_response_callback(self, callback_function):
        with self.callbacks_lock:
            self.response_callbacks.remove(callback_function)

    def flush_response_callbacks(self):
        with self.callbacks_lock:
            for callback in self.response_callbacks:
                try:
                    self.response_callbacks.remove(callback)
                except ValueError:
                    pass
