# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)


import time
import logging
import threading
import queue
from PySide2 import QtCore

import log

class PrinterModel(QtCore.QObject):

    SLEEP_ON_PRINTER_INTERFACE_FAIL = 0.5 

    UPDATE_LOOP_TIME = 1
    TRANSMIT_CALLBACKS_LOOP_TIME = 0.1

    QUEUE_SIZE = 1000

    X_OFFSET_HIGH = 40
    X_OFFSET_LOW = 60

    Y_OFFSET_HIGH = 10.0
    Y_OFFSET_LOW = -10.0

    Z_OFFSET_HIGH = 0
    Z_OFFSET_LOW = -2.0


    # QProperty signals
    on_token = QtCore.Signal()
    on_firstExtruderTemp = QtCore.Signal()
    on_secondExtruderTemp = QtCore.Signal()
    on_heatBedTemp = QtCore.Signal()
    on_chamberTemp = QtCore.Signal()
    on_firstExtruderTargetTemp = QtCore.Signal()
    on_secondExtruderTargetTemp = QtCore.Signal()
    on_heatBedTargetTemp = QtCore.Signal()
    on_chamberTargetTemp = QtCore.Signal()
    on_currentGCodeLine = QtCore.Signal()
    on_totalGCodeLine = QtCore.Signal()
    on_percent = QtCore.Signal()
    on_xOffset = QtCore.Signal()
    on_yOffset = QtCore.Signal()
    on_zOffset = QtCore.Signal()
    on_joystickFeedrate = QtCore.Signal()
    on_printerError = QtCore.Signal()
    on_authToken = QtCore.Signal()
    on_responseCallback = QtCore.Signal(str, bool)

    def __init__(self, core_model):
        super().__init__()
        self.logger = logging.getLogger(__name__)
        self._coreModel = core_model
        self._app = core_model._app
        self._firstExtruderTemp = 0
        self._secondExtruderTemp = 0
        self._heatBedTemp = 0
        self._chamberTemp = 0
        self._firstExtruderTargetTemp = 0
        self._secondExtruderTargetTemp = 0
        self._heatBedTargetTemp = 0
        self._chamberTargetTemp= 0
        self._percent = 0
        self._running = False
        self._token = None
        self._currentGCodeLine = 0
        self._totalGCodeLine = 0
        self._getPrinterInterfaFailsCount = 0
        self._joystickFeedrate = 300
        self._printerError = ""
        self._authToken = ""
        self._responsesQueue = queue.SimpleQueue()
        #self._xOffset = config.get_settings()['calibration']['x_offset']
        #self._yOffset = config.get_settings()['calibration']['y_offset']
        #self._zOffset = config.get_settings()['calibration']['z_offset']
        self._printerUpdateThread = threading.Thread(target=self._printerUpdateLoop)
        self._printerUpdateThread.start()
        self._responseCallbacksThread = threading.Thread(target=self._responseCallbacksLoop)
        self._responseCallbacksThread.start()

    def _getPrinterInterface(self):
        try:
            pi = self._app.printer_interfaces[0]
            return pi
        except (IndexError, AttributeError):
            message = "Warning: no printer interface"
            self.logger.warning(message)
            time.sleep(self.SLEEP_ON_PRINTER_INTERFACE_FAIL)
            if self._app.stop_flag:
                raise RuntimeError(message) from None
            return self._getPrinterInterface()

    @log.log_exception
    def _printerUpdateLoop(self):
        while not self._app.stop_flag:
            try:
                printer_interface = self._getPrinterInterface()
                if printer_interface:
                    self._authToken = printer_interface.printer_token
                    self.on_authToken.emit()
                    state_report = printer_interface.state_report()
                    state = state_report['state']
                    if state not in ('connecting', 'closing', 'error'):
                        temps = list(state_report['temps'])
                        while len(temps) <= 4:
                            temps.append(0)
                        self._heatBedTemp = int(temps[0])
                        self._firstExtruderTemp = int(temps[1])
                        self._secondExtruderTemp = int(temps[2])
                        self._chamberTemp = int(temps[3])
                        # TODO refactor this copypaste crap
                        for signal in (self.on_firstExtruderTemp,
                                        self.on_secondExtruderTemp,
                                        self.on_heatBedTemp,
                                        self.on_chamberTemp):
                            signal.emit()
                        target_temps = list(state_report['target_temps'])
                        while len(target_temps) <= 4:
                            target_temps.append(0)
                        self._heatBedTargetTemp = int(target_temps[0])
                        self._firstExtruderTargetTemp = int(target_temps[1])
                        self._secondExtruderTargetTemp = int(target_temps[2])
                        self._chamberTargetTemp = int(target_temps[3])
                        for signal in (self.on_firstExtruderTargetTemp,
                                        self.on_secondExtruderTargetTemp,
                                        self.on_heatBedTargetTemp,
                                        self.on_chamberTargetTemp):
                            signal.emit()
                        self._percent = state_report['percent']
                        self._currentGCodeLine = state_report['line_number']
                        self._totalGCodeLine = printer_interface.printer.total_gcodes
                        self.on_currentGCodeLine.emit()
                        self.on_totalGCodeLine.emit()
                        self.on_percent.emit()
                    if state == 'printing':
                        state = self._coreModel.STATE_PRINTING
                    elif state == 'downloading':
                        state = self._coreModel.STATE_DOWNLOADING
                    elif state == 'paused':
                        state = self._coreModel.STATE_PAUSED
                    elif state == 'ready' or state == 'local_mode':
                        state = self._coreModel.STATE_IDLE
                    else:
                        #self._coreModel.showMessage.emit("error", "message", "Continue")
                        code = printer_interface.registration_code
                        if state == 'error':
                            self.on_printerError.emit()
                            if printer_interface:
                                for error in getattr(printer_interface, "errors", []):
                                    self._coreModel.showMessage.emit("error", error["message"], "Continue")
                        elif code:
                            self._token = code
                            self.on_token.emit()
                            state = self._coreModel.STATE_REGISTRATION
                        elif state == 'closing' or state == 'connecting':
                            state = self._coreModel.STATE_CONNECTING
                    self._coreModel.setState(state)
            except:
                self.logger.exception("Exception in updateLoop:")
            finally:
                time.sleep(self.UPDATE_LOOP_TIME)

    @log.log_exception
    def _responseCallbacksLoop(self):
        while not self._app.stop_flag:
            try:
                response = self._responsesQueue.get(block=True, timeout=5)
            except queue.Empty:
                time.sleep(self.TRANSMIT_CALLBACKS_LOOP_TIME)
            else:
                self.on_responseCallback.emit(*response)

    def loadGcodes(self, gcodes, print_file=True):
        while not self._app.stop_flag:
            pi = self._getPrinterInterface()
            if pi:
                printer = pi.printer
                if printer:
                    if print_file:
                        printer.load_gcodes(gcodes) 
                    else:
                        printer.unbuffered_gcodes(gcodes) 
                    return True
            else:
                self.logger.error("No printer to execute gcodes")
            time.sleep(self.UPDATE_LOOP_TIME)

    def addToResponsesQueue(self, line, success):
        try:
            self._responsesQueue.put_nowait((line, success))
        except queue.Full:
            self.logger.warning("Full gcode response queue")

    @QtCore.Slot()
    def disconnect_printer(self):
        self._getPrinterInterface().close()

    @QtCore.Slot()
    def cancelPrint(self):
        self._getPrinterInterface().cancel_locally()

    @QtCore.Slot()
    def pausePrint(self):
        self._getPrinterInterface().printer.pause()

    @QtCore.Slot()
    def resumePrint(self):
        self._getPrinterInterface().printer.unpause()

    @QtCore.Slot(str)
    def executeGcodeLine(self, gcode_line):
        self.logger.info("Executing gcodes: " + str(gcode_line))
        if gcode_line:
            if not gcode_line.endswith("\n"):
                gcode_line += "\n"
            self.loadGcodes(gcode_line.encode("ascii"), print_file=False)

    @QtCore.Slot()
    def enableResponseCallbacks(self):
        try:
            self._getPrinterInterface().printer.add_response_callback(self.addToResponsesQueue)
        except:
            self.logger.exception() #TODO comment out in production
            pass

    @QtCore.Slot()
    def disableResponseCallbacks(self):
        try:
            self._getPrinterInterface().printer.flush_response_callbacks()
        except:
            self.logger.exception() #TODO comment out in production
            pass

    @QtCore.Slot()
    def localModeOn(self):
        self._getPrinterInterface().turn_on_local_mode()

    @QtCore.Slot()
    def localModeOff(self):
        self._getPrinterInterface().turn_off_local_mode()

    @QtCore.Slot()
    def resetBoard(self):
        success = False
        while not success and not self.stop_flag:
            wait = 5
            try:
                success = self._getPrinterInterface().printer.reset()
            except:
                wait = 0.1
            if success:
                break
            time.sleep(wait)

   # @QtCore.Property('QVariantList', notify=on_temperature)
   # def temperaturesList(self):
   #     return self._temperatureList

    @QtCore.Property(int, notify=on_firstExtruderTemp)
    def firstExtruderTemp(self):
        return self._firstExtruderTemp

    @QtCore.Property(int, notify=on_secondExtruderTemp)
    def secondExtruderTemp(self):
        return self._secondExtruderTemp

    @QtCore.Property(int, notify=on_heatBedTemp)
    def heatBedTemp(self):
        return self._heatBedTemp

    @QtCore.Property(int, notify=on_chamberTemp)
    def chamberTemp(self):
        return self._chamberTemp

    @QtCore.Property(int, notify=on_firstExtruderTargetTemp)
    def firstExtruderTargetTemp(self):
        return self._firstExtruderTargetTemp

    @QtCore.Property(int, notify=on_secondExtruderTargetTemp)
    def secondExtruderTargetTemp(self):
        return self._secondExtruderTargetTemp

    @QtCore.Property(int, notify=on_heatBedTargetTemp)
    def heatBedTargetTemp(self):
        return self._heatBedTargetTemp

    @QtCore.Property(int, notify=on_chamberTargetTemp)
    def chamberTargetTemp(self):
        return self._chamberTargetTemp

    @QtCore.Property(int, notify=on_percent)
    def percent(self):
        return self._percent

    @QtCore.Property(str, notify=on_token)
    def token(self):
        return self._token

    @QtCore.Property(int, notify=on_currentGCodeLine)
    def currentGCodeLine(self):
        return self._currentGCodeLine

    @QtCore.Property(int, notify=on_totalGCodeLine) #TODO rename ...Line to Lines
    def totalGCodeLine(self):
        return self._totalGCodeLine

    @QtCore.Property(int, notify=on_joystickFeedrate)
    def joystickFeedrate(self):
        return self._joystickFeedrate

    @QtCore.Property(str, notify=on_printerError)
    def printerError(self):
        return self._printerError

    @QtCore.Property(bool, notify=on_authToken)
    def gotAuthToken(self):
        return bool(self._authToken)
